// Links an AmigaDOS object file and outputs a .BIN file with the
// PlusShell format.
// Written by Rusty Wagner for PlusShell

#include "progstub.h"
#include "progbss.h"
#include "libstub.h"
#include "fargstub.h"

#define PROGRAM 0
#define LIBRARY 1

#define TI92 1
#define TI89 2

unsigned char buf[65536];
int relocSize=0;
unsigned short relocBuf[32768];
char str[64];
int funcCount=0;
unsigned short funcOfs[1024];
int mainOfs,commentOfs;
int stubLen,stubStart;
int libCount=0;
unsigned char libName[32][8];
int libRelCount[32];
unsigned short libRelOfs[32][2048];
unsigned short libRelNum[32][2048];
int romCount=0;
unsigned short romOfs[16384];
unsigned short romNum[16384];
int ramCount=0;
int ramRelocSize=2;
unsigned short ramOfs[16384];
unsigned short ramNum[16384];
unsigned long ramDisp[16384];
int bssSize=0;
int bssRelocSize=0;
int bssRelocBuf[16384];
int fargoConvert=0;

int output89=0;
int output92plus=0;

unsigned long ReadDWord(FILE *fp)
{
        unsigned long val;
        val=fgetc(fp)<<24;
        val|=fgetc(fp)<<16;
        val|=fgetc(fp)<<8;
        val|=fgetc(fp);
        return val;
}

int GetLibNum(char *str)
{
        unsigned char buf[8];
        int i,j,find;

        for (i=0;i<8;i++) buf[i]=0;
        for (i=0;str[i]!='@';i++) buf[i]=str[i];
        for (find=-1,i=0;i<libCount;i++)
        {
                for (j=0;j<8;j++)
                {
                        if (buf[j]!=libName[i][j])
                                break;
                }
                if (j==8)
                {
                        find=i;
                        break;
                }
        }
        if (find!=-1)
                return find;
        for (i=0;i<8;i++) libName[libCount][i]=buf[i];
        libRelCount[libCount]=0;
        libCount++;
        return libCount-1;
}

int GetLibFunc(char *str)
{
        int i;

        sscanf(&str[strlen(str)-4],"%X",&i);
        return i;
}

void SetDWord(unsigned char *buf,int ofs,int v)
{
        buf[ofs]=(v>>24)&0xff;
        buf[ofs+1]=(v>>16)&0xff;
        buf[ofs+2]=(v>>8)&0xff;
        buf[ofs+3]=v&0xff;
}

void SetWord(unsigned char *buf,int ofs,int v)
{
        buf[ofs]=(v>>8)&0xff;
        buf[ofs+1]=v&0xff;
}

void AddDWord(unsigned char *buf,int ofs,int n)
{
        int v;

        v=buf[ofs]<<24;
        v|=buf[ofs+1]<<16;
        v|=buf[ofs+2]<<8;
        v|=buf[ofs+3];
        v+=n;
        buf[ofs]=(v>>24)&0xff;
        buf[ofs+1]=(v>>16)&0xff;
        buf[ofs+2]=(v>>8)&0xff;
        buf[ofs+3]=v&0xff;
}

void AddWord(unsigned char *buf,int ofs,int n)
{
        int v;

        v=buf[ofs]<<8;
        v|=buf[ofs+1];
        v+=n;
        buf[ofs]=(v>>8)&0xff;
        buf[ofs+1]=v&0xff;
}

int makebin(char *name)
{
        FILE *fp,*out;
        char str[256];
        int len,i,j,ofs,type,k,lib,fnc,n,t,v,stubEnable,calcs;

        sprintf(str,"%s.o",name);
        fp=fopen(str,"rb");
        sprintf(str,"%s.bin",name);
        out=fopen(str,"wb");
        fseek(fp,0x18,SEEK_SET);
        len=ReadDWord(fp)<<2;
        for (i=0;i<len;i++)
                buf[i]=fgetc(fp);
        k=ReadDWord(fp);
        relocSize=0;
        printf("Reading relocation table...\n");
        if (k==0x3ec)
        {
                while (1)
                {
                        n=ReadDWord(fp);
                        if (n==0)
                        {
                                k=ReadDWord(fp);
                                break;
                        }
                        t=ReadDWord(fp);
                        if (t==0)
                        {
                                i=relocSize;
                                relocSize+=n<<1;
                                for (;i<relocSize;i+=2)
                                {
                                        ofs=ReadDWord(fp);
                                        relocBuf[i]=(buf[ofs+2]<<8)+buf[ofs+3];
                                        relocBuf[i+1]=ofs;
                                }
                        }
                        else
                        {
                                i=bssRelocSize;
                                bssRelocSize=n<<1;
                                for (;i<bssRelocSize;i+=2)
                                {
                                        ofs=ReadDWord(fp);
                                        bssRelocBuf[i]=ofs;
                                        bssRelocBuf[i+1]=(buf[ofs+2]<<8)+buf[ofs+3];
                                }
                        }
                }
        }
        else
                relocSize=0;
        type=PROGRAM;
        libCount=0;
        romCount=0;
        mainOfs=0;
        commentOfs=-1;
        stubEnable=1;
        calcs=0;
        printf("Reading symbol table...\n");
        if (k==0x3ef)
        {
                funcCount=0;
                while (1)
                {
                        k=ReadDWord(fp);
                        if (!(k&0x7f000000)) break;
                        if ((k&0xff000000)==0x81000000)
                        {
                                k&=0xffff;
                                for (i=0;k;k--)
                                {
                                        str[i++]=fgetc(fp);
                                        str[i++]=fgetc(fp);
                                        str[i++]=fgetc(fp);
                                        str[i++]=fgetc(fp);
                                }
                                str[i++]=0;
                                if (!strncmp(str,"_ROM_CALL_",10))
                                {
                                        sscanf(&str[10],"%X",&j);
                                        k=ReadDWord(fp);
                                        for (i=0;i<k;i++)
                                        {
                                                romOfs[romCount]=ReadDWord(fp);
                                                romNum[romCount++]=j;
                                        }
                                        continue;
                                }
                                if (!strcmp(str,"tios::kb_globals"))
                                {
                                        k=ReadDWord(fp);
                                        for (i=0;i<k;i++)
                                        {
                                                ofs=ReadDWord(fp);
                                                v=buf[ofs]<<24;
                                                v|=buf[ofs+1]<<16;
                                                v|=buf[ofs+2]<<8;
                                                v|=buf[ofs+3];
                                                if (v==0x1c)
                                                {
                                                        ramOfs[ramCount]=ofs;
                                                        ramNum[ramCount++]=4;
                                                        ramRelocSize+=4;
                                                }
                                                else if (v==0x1e)
                                                {
                                                        ramOfs[ramCount]=ofs;
                                                        ramNum[ramCount++]=5;
                                                        ramRelocSize+=4;
                                                }
                                                else if ((v==0xc)||(v==-6))
                                                {
                                                        if (calcs&TI89)
                                                        {
                                                                printf("Symbol 'tios::kb_globals+%X' is currently undefined on TI-89\n",v);
                                                                return 1;
                                                        }
                                                        if (v==0xc)
                                                                SetDWord(buf,ofs,0x75a2);
                                                        if (v==-6)
                                                                SetDWord(buf,ofs,0x75be);
                                                }
                                                else
                                                {
                                                        printf("Undefined symbol 'tios::kb_globals+%X'\n",v);
                                                        return 1;
                                                }
                                        }
                                        continue;
                                }
                                if (!strcmp(str,"FOLDER_LIST_HANDLE"))
                                {
                                        k=ReadDWord(fp);
                                        for (i=0;i<k;i++)
                                        {
                                                ofs=ReadDWord(fp);
                                                AddDWord(buf,ofs,8);
                                        }
                                        continue;
                                }
                                if (!strcmp(str,"LCD_MEM"))
                                {
                                        k=ReadDWord(fp);
                                        for (i=0;i<k;i++)
                                        {
                                                ofs=ReadDWord(fp);
                                                AddDWord(buf,ofs,0x4c00);
                                        }
                                        continue;
                                }
                                if (!strcmp(str,"LCD_WIDTH"))
                                {
                                        k=ReadDWord(fp);
                                        for (i=0;i<k;i++)
                                        {
                                                ofs=ReadDWord(fp);
                                                v=buf[ofs]<<24;
                                                v|=buf[ofs+1]<<16;
                                                v|=buf[ofs+2]<<8;
                                                v|=buf[ofs+3];
                                                ramOfs[ramCount]=ofs;
                                                ramNum[ramCount]=0x4001;
                                                ramDisp[ramCount++]=v;
                                                ramRelocSize+=8;
                                        }
                                        continue;
                                }
                                if (!strcmp(str,"LCD_HEIGHT"))
                                {
                                        k=ReadDWord(fp);
                                        for (i=0;i<k;i++)
                                        {
                                                ofs=ReadDWord(fp);
                                                v=buf[ofs]<<24;
                                                v|=buf[ofs+1]<<16;
                                                v|=buf[ofs+2]<<8;
                                                v|=buf[ofs+3];
                                                ramOfs[ramCount]=ofs;
                                                ramNum[ramCount]=0x4002;
                                                ramDisp[ramCount++]=v;
                                                ramRelocSize+=8;
                                        }
                                        continue;
                                }
                                if (!strcmp(str,"APD_INIT"))
                                {
                                        k=ReadDWord(fp);
                                        for (i=0;i<k;i++)
                                        {
                                                ofs=ReadDWord(fp);
                                                AddDWord(buf,ofs,0x5b10);
                                        }
                                        continue;
                                }
                                if (!strcmp(str,"APD_TIMER"))
                                {
                                        k=ReadDWord(fp);
                                        for (i=0;i<k;i++)
                                        {
                                                ofs=ReadDWord(fp);
                                                AddDWord(buf,ofs,0x5b14);
                                        }
                                        continue;
                                }
                                if (!strcmp(str,"APD_FLAG"))
                                {
                                        k=ReadDWord(fp);
                                        for (i=0;i<k;i++)
                                        {
                                                ofs=ReadDWord(fp);
                                                AddDWord(buf,ofs,0x5b42);
                                        }
                                        continue;
                                }
                                if (!strcmp(str,"tios::Heap"))
                                {
                                        k=ReadDWord(fp);
                                        for (i=0;i<k;i++)
                                        {
                                                ofs=ReadDWord(fp);
                                                ramOfs[ramCount]=ofs;
                                                ramNum[ramCount++]=3;
                                                ramRelocSize+=4;
                                        }
                                        continue;
                                }
                                if (!strcmp(str,"CALCULATOR"))
                                {
                                        k=ReadDWord(fp);
                                        for (i=0;i<k;i++)
                                        {
                                                ofs=ReadDWord(fp);
                                                ramOfs[ramCount]=ofs;
                                                ramNum[ramCount++]=0;
                                                ramRelocSize+=4;
                                        }
                                        continue;
                                }
                                if (!strcmp(str,"SYM_ENTRY.name"))
                                {
                                        k=ReadDWord(fp);
                                        for (i=0;i<k;i++)
                                        {
                                                ofs=ReadDWord(fp);
                                                SetDWord(buf,ofs,0);
                                        }
                                        continue;
                                }
                                if (!strcmp(str,"SYM_ENTRY.flags"))
                                {
                                        k=ReadDWord(fp);
                                        for (i=0;i<k;i++)
                                        {
                                                ofs=ReadDWord(fp);
                                                SetDWord(buf,ofs,10);
                                        }
                                        continue;
                                }
                                if (!strcmp(str,"SYM_ENTRY.flags32"))
                                {
                                        k=ReadDWord(fp);
                                        for (i=0;i<k;i++)
                                        {
                                                ofs=ReadDWord(fp);
                                                SetDWord(buf,ofs,8);
                                        }
                                        continue;
                                }
                                if (!strcmp(str,"SYM_ENTRY.hVal"))
                                {
                                        k=ReadDWord(fp);
                                        for (i=0;i<k;i++)
                                        {
                                                ofs=ReadDWord(fp);
                                                SetDWord(buf,ofs,12);
                                        }
                                        continue;
                                }
                                if (!strcmp(str,"SYM_ENTRY_LENGTH"))
                                {
                                        k=ReadDWord(fp);
                                        for (i=0;i<k;i++)
                                        {
                                                ofs=ReadDWord(fp);
                                                SetDWord(buf,ofs,14);
                                        }
                                        continue;
                                }
                                if (strlen(str)<=4)
                                {
                                        printf("Undefined symbol '%s'\n",str);
                                        return 1;
                                }
                                if (str[strlen(str)-5]!='@')
                                {
                                        printf("Undefined symbol '%s'\n",str);
                                        return 1;
                                }
                                lib=GetLibNum(str);
                                fnc=GetLibFunc(str);
                                k=ReadDWord(fp);
                                for (i=0;i<k;i++)
                                {
                                        libRelOfs[lib][libRelCount[lib]]=ReadDWord(fp);
                                        libRelNum[lib][libRelCount[lib]]=fnc;
                                        libRelCount[lib]++;
                                }
                        }
                        else if ((k&0xff000000)==0x8a000000)
                        {
                                k&=0xffff;
                                for (i=0;k;k--)
                                {
                                        str[i++]=fgetc(fp);
                                        str[i++]=fgetc(fp);
                                        str[i++]=fgetc(fp);
                                        str[i++]=fgetc(fp);
                                }
                                str[i++]=0;
                                if (!strcmp(str,"LCD_WIDTH"))
                                {
                                        k=ReadDWord(fp);
                                        for (i=0;i<k;i++)
                                        {
                                                ofs=ReadDWord(fp);
                                                v=buf[ofs]<<8;
                                                v|=buf[ofs+1];
                                                ramOfs[ramCount]=ofs;
                                                ramNum[ramCount]=0xc001;
                                                ramDisp[ramCount++]=v;
                                                ramRelocSize+=6;
                                        }
                                        continue;
                                }
                                if (!strcmp(str,"LCD_HEIGHT"))
                                {
                                        k=ReadDWord(fp);
                                        for (i=0;i<k;i++)
                                        {
                                                ofs=ReadDWord(fp);
                                                v=buf[ofs]<<8;
                                                v|=buf[ofs+1];
                                                ramOfs[ramCount]=ofs;
                                                ramNum[ramCount]=0xc002;
                                                ramDisp[ramCount++]=v;
                                                ramRelocSize+=6;
                                        }
                                        continue;
                                }
                                if (!strcmp(str,"FOLDER_LIST_HANDLE"))
                                {
                                        k=ReadDWord(fp);
                                        for (i=0;i<k;i++)
                                        {
                                                ofs=ReadDWord(fp);
                                                AddWord(buf,ofs,8);
                                        }
                                        continue;
                                }
                                if (!strcmp(str,"SYM_ENTRY.name"))
                                {
                                        k=ReadDWord(fp);
                                        for (i=0;i<k;i++)
                                        {
                                                ofs=ReadDWord(fp);
                                                SetWord(buf,ofs,0);
                                        }
                                        continue;
                                }
                                if (!strcmp(str,"SYM_ENTRY.flags"))
                                {
                                        k=ReadDWord(fp);
                                        for (i=0;i<k;i++)
                                        {
                                                ofs=ReadDWord(fp);
                                                SetWord(buf,ofs,10);
                                        }
                                        continue;
                                }
                                if (!strcmp(str,"SYM_ENTRY.flags32"))
                                {
                                        k=ReadDWord(fp);
                                        for (i=0;i<k;i++)
                                        {
                                                ofs=ReadDWord(fp);
                                                SetWord(buf,ofs,8);
                                        }
                                        continue;
                                }
                                if (!strcmp(str,"SYM_ENTRY.hVal"))
                                {
                                        k=ReadDWord(fp);
                                        for (i=0;i<k;i++)
                                        {
                                                ofs=ReadDWord(fp);
                                                SetWord(buf,ofs,12);
                                        }
                                        continue;
                                }
                                if (!strcmp(str,"SYM_ENTRY_LENGTH"))
                                {
                                        k=ReadDWord(fp);
                                        for (i=0;i<k;i++)
                                        {
                                                ofs=ReadDWord(fp);
                                                SetWord(buf,ofs,14);
                                        }
                                        continue;
                                }
                                printf("%s can not be used as word value\n",str);
                                return 1;
                        }
                        else
                        {
                                k&=0xffff;
                                for (i=0;k;k--)
                                {
                                        str[i++]=fgetc(fp);
                                        str[i++]=fgetc(fp);
                                        str[i++]=fgetc(fp);
                                        str[i++]=fgetc(fp);
                                }
                                str[i++]=0;
                                ofs=ReadDWord(fp);
                                if (!strcmp(str,"_library"))
                                        type=LIBRARY;
                                else if (!strcmp(str,"_main"))
                                        mainOfs=ofs;
                                else if (!strcmp(str,"_comment"))
                                        commentOfs=ofs;
                                else if (!strcmp(str,"_FARGO_CONVERT"))
                                        fargoConvert=1;
                                else if (!strcmp(str,"_nostub"))
                                        stubEnable=0;
                                else if (!strcmp(str,"_ti89"))
                                        calcs|=TI89;
                                else if (!strcmp(str,"_ti92plus"))
                                        calcs|=TI92;
                                else if (strlen(str)>4)
                                {
                                        if (str[strlen(str)-5]=='@')
                                                funcOfs[funcCount++]=ofs;
                                }
                        }
                }
                k=ReadDWord(fp);
        }
        if (calcs==0) calcs=TI92;
        if ((!stubEnable)&&(type==LIBRARY))
                stubEnable=1;
        if (fargoConvert)
        {
                lib=GetLibNum("util@0011");
                libRelOfs[lib][libRelCount[lib]]=-22;
                libRelNum[lib][libRelCount[lib]]=0x11;
                libRelCount[lib]++;
                libRelOfs[lib][libRelCount[lib]]=-12;
                libRelNum[lib][libRelCount[lib]]=0x12;
                libRelCount[lib]++;
        }
        if ((!stubEnable)&&(romCount))
        {
                printf("Programs without stubs can not have direct ROM calls\n");
                return 1;
        }
        if ((!stubEnable)&&(libCount))
        {
                printf("Programs without stubs can not have library calls\n");
                return 1;
        }
        while (!feof(fp))
        {
                if (k==0x3eb)
                {
                        bssSize=ReadDWord(fp)<<2;
                        k=ReadDWord(fp);
                }
                else if (k==0x3e8)
                {
                        ReadDWord(fp);
                        ReadDWord(fp);
                        k=ReadDWord(fp);
                }
                else
                        k=ReadDWord(fp);
        }
        if ((!stubEnable)&&(bssSize))
        {
                printf("Programs without stubs can not have BSS sections\n");
                return 1;
        }
        printf("Writing program...\n");
        if (type==PROGRAM)
        {
                if (bssSize>0) progStubLen=progBssStubLen;
                if (!stubEnable) progStubLen=0;
                stubLen=24+progStubLen+(libCount<<3)+(romCount<<2)+(bssRelocSize<<1)+(fargoConvert*fargoStubLen)+ramRelocSize;
                for (i=0;i<libCount;i++)
                {
                        stubLen+=2;
                        stubLen+=libRelCount[i]<<2;
                }
                stubStart=stubLen-4-progStubLen-(fargoConvert*fargoStubLen);
                fputc(0x60,out); fputc(0,out);
                fputc(((stubStart-2)>>8)&0xff,out);
                fputc((stubStart-2)&0xff,out);
                fputc('P',out); fputc('S',out); fputc('v',out); fputc('1',out);
                if (calcs==TI92)
                        fputc(0,out);
                else if (calcs==TI89)
                        fputc(0xc0,out);
                else // Both TI92 and TI89
                        fputc(0x80,out);
                fputc('P',out);
                if (commentOfs==-1) commentOfs=-stubLen;
                fputc(((commentOfs+stubLen)>>8)&0xff,out);
                fputc((commentOfs+stubLen)&0xff,out);
                fputc((romCount>>8)&0xff,out); fputc(romCount&0xff,out);
                for (i=0;i<romCount;i++)
                {
                        romOfs[i]+=stubLen;
                        fputc((romOfs[i]>>8)&0xff,out);
                        fputc(romOfs[i]&0xff,out);
                        fputc((romNum[i]>>8)&0xff,out);
                        fputc(romNum[i]&0xff,out);
                }
                fputc((ramCount>>8)&0xff,out); fputc(ramCount&0xff,out);
                for (i=0;i<ramCount;i++)
                {
                        ramOfs[i]+=stubLen;
                        fputc((ramOfs[i]>>8)&0xff,out);
                        fputc(ramOfs[i]&0xff,out);
                        fputc((ramNum[i]>>8)&0xff,out);
                        fputc(ramNum[i]&0xff,out);
                        if (ramNum[i]&0x4000)
                        {
                                if (ramNum[i]&0x8000)
                                {
                                        fputc((ramDisp[i]>>8)&0xff,out);
                                        fputc(ramDisp[i]&0xff,out);
                                }
                                else
                                {
                                        fputc((ramDisp[i]>>24)&0xff,out);
                                        fputc((ramDisp[i]>>16)&0xff,out);
                                        fputc((ramDisp[i]>>8)&0xff,out);
                                        fputc(ramDisp[i]&0xff,out);
                                }
                        }
                }
                fputc((libCount>>8)&0xff,out); fputc(libCount&0xff,out);
                for (i=0;i<libCount;i++)
                {
                        for (j=0;j<8;j++)
                                fputc(libName[i][j],out);
                        fputc((libRelCount[i]>>8)&0xff,out);
                        fputc(libRelCount[i]&0xff,out);
                        for (j=0;j<libRelCount[i];j++)
                        {
                                libRelOfs[i][j]+=stubLen;
                                fputc((libRelOfs[i][j]>>8)&0xff,out);
                                fputc(libRelOfs[i][j]&0xff,out);
                                fputc((libRelNum[i][j]>>8)&0xff,out);
                                fputc(libRelNum[i][j]&0xff,out);
                        }
                }
                fputc((bssSize>>8)&0xff,out); fputc(bssSize&0xff,out);
                fputc((bssRelocSize>>9)&0xff,out);
                fputc((bssRelocSize>>1)&0xff,out);
                for (i=0;i<bssRelocSize;i++)
                {
                        if ((i&1)==0)
                                bssRelocBuf[i]+=stubLen;
                        fputc((bssRelocBuf[i]>>8)&0xff,out);
                        fputc(bssRelocBuf[i]&0xff,out);
                }
                if (bssSize==0)
                {
                        for (i=0;i<progStubLen;i++)
                                fputc(progStub[i],out);
                }
                else
                {
                        for (i=0;i<progStubLen;i++)
                                fputc(progBssStub[i],out);
                }
                if (fargoConvert)
                {
                        for (i=0;i<fargoStubLen;i++)
                                fputc(fargoStub[i],out);
                }
                fputc(0x60,out); fputc(0,out);
                fputc(((mainOfs+2)>>8)&0xff,out);
                fputc((mainOfs+2)&0xff,out);
                for (i=0;i<len;i++)
                        fputc(buf[i],out);
                fputc(0,out); fputc(0,out);
                for (i=0;i<relocSize;i++)
                {
                        relocBuf[i]+=stubLen;
                        fputc((relocBuf[i]>>8)&0xff,out);
                        fputc(relocBuf[i]&0xff,out);
                }
                fputc(0,out); fputc(0,out);
                if (stubEnable)
                {
                        fputc(((stubStart+6)>>8)&0xff,out);
                        fputc((stubStart+6)&0xff,out);
                }
        }
        else
        {
                if (bssSize>0)
                {
                        printf("Error: Libraries cannot have BSS sections\n");
                        return 1;
                }
                stubLen=24+libStubSize+(funcCount<<1)+(romCount<<2)+(libCount<<3)+(bssRelocSize<<1)+ramRelocSize;
                for (i=0;i<libCount;i++)
                {
                        stubLen+=2;
                        stubLen+=libRelCount[i]<<2;
                }
                fputc(0x60,out); fputc(0,out);
                fputc(((stubLen-2-libStubSize)>>8)&0xff,out);
                fputc((stubLen-2-libStubSize)&0xff,out);
                fputc('P',out); fputc('S',out); fputc('v',out); fputc('1',out);
                if (calcs==TI92)
                        fputc(0,out);
                else if (calcs==TI89)
                        fputc(0xc0,out);
                else // Both TI92 and TI89
                        fputc(0x80,out);
                fputc('L',out);
                fputc(0x60,out); fputc(0,out);
                fputc(((stubLen-16)>>8)&0xff,out);
                fputc((stubLen-16)&0xff,out);
                fputc((funcCount>>8)&0xff,out);
                fputc(funcCount&0xff,out);
                for (i=0;i<funcCount;i++)
                {
                        fputc(((funcOfs[i]+stubLen)>>8)&0xff,out);
                        fputc((funcOfs[i]+stubLen)&0xff,out);
                }
                fputc((romCount>>8)&0xff,out); fputc(romCount&0xff,out);
                for (i=0;i<romCount;i++)
                {
                        romOfs[i]+=stubLen;
                        fputc((romOfs[i]>>8)&0xff,out);
                        fputc(romOfs[i]&0xff,out);
                        fputc((romNum[i]>>8)&0xff,out);
                        fputc(romNum[i]&0xff,out);
                }
                fputc((ramCount>>8)&0xff,out); fputc(ramCount&0xff,out);
                for (i=0;i<ramCount;i++)
                {
                        ramOfs[i]+=stubLen;
                        fputc((ramOfs[i]>>8)&0xff,out);
                        fputc(ramOfs[i]&0xff,out);
                        fputc((ramNum[i]>>8)&0xff,out);
                        fputc(ramNum[i]&0xff,out);
                        if (ramNum[i]&0x4000)
                        {
                                if (ramNum[i]&0x8000)
                                {
                                        fputc((ramDisp[i]>>8)&0xff,out);
                                        fputc(ramDisp[i]&0xff,out);
                                }
                                else
                                {
                                        fputc((ramDisp[i]>>24)&0xff,out);
                                        fputc((ramDisp[i]>>16)&0xff,out);
                                        fputc((ramDisp[i]>>8)&0xff,out);
                                        fputc(ramDisp[i]&0xff,out);
                                }
                        }
                }
                fputc((libCount>>8)&0xff,out); fputc(libCount&0xff,out);
                for (i=0;i<libCount;i++)
                {
                        for (j=0;j<8;j++)
                                fputc(libName[i][j],out);
                        fputc((libRelCount[i]>>8)&0xff,out);
                        fputc(libRelCount[i]&0xff,out);
                        for (j=0;j<libRelCount[i];j++)
                        {
                                libRelOfs[i][j]+=stubLen;
                                fputc((libRelOfs[i][j]>>8)&0xff,out);
                                fputc(libRelOfs[i][j]&0xff,out);
                                fputc((libRelNum[i][j]>>8)&0xff,out);
                                fputc(libRelNum[i][j]&0xff,out);
                        }
                }
                fputc((bssSize>>8)&0xff,out); fputc(bssSize&0xff,out);
                fputc((bssRelocSize>>8)&0xff,out);
                fputc(bssRelocSize&0xff,out);
                for (i=0;i<bssRelocSize;i++)
                {
                        fputc((bssRelocBuf[i]>>8)&0xff,out);
                        fputc(bssRelocBuf[i]&0xff,out);
                }
                for (i=0;i<libStubSize;i++)
                        fputc(libStub[i],out);
                for (i=0;i<len;i++)
                        fputc(buf[i],out);
                fputc(0,out); fputc(0,out);
                for (i=0;i<relocSize;i++)
                {
                        relocBuf[i]+=stubLen;
                        fputc((relocBuf[i]>>8)&0xff,out);
                        fputc(relocBuf[i]&0xff,out);
                }
        }
        fclose(fp); fclose(out);
        if (calcs&TI89) output89=1;
        if (calcs&TI92) output92plus=1;
        return 0;
}

